package bst;

import java.util.Comparator;

/**
 * @author Aapo Kyrola, Sulake Corporation 2008
 */

public class BinarySearchTree<K, T> {

    private Node root;
    private Comparator<K> comp;
    private int maxDepth = 0, size = 0;


    public BinarySearchTree(Comparator<K> comp) {
        this.comp = comp;
        root = null;
    }

    public Node getRoot() {
        return root;
    }

    public int getMaxDepth() {
        return maxDepth;
    }

    public int size() {
        return size;
    }

    public void insert(K key, T value) {
        Node nd = new Node();
        nd.key = key;
        nd.payload = value;

        if (root == null) {
            root = nd;
            nd.depth = 0;
        } else {
            Node y = null;
            Node x = root;
            while(x != null) {
                y = x;
                if (comp.compare(key, x.key) > 0) {
                    x = x.right;
                } else x = x.left;
            }
            nd.p = y;
            nd.depth = nd.p.depth + 1;
            if (comp.compare(key, y.key) > 0) {
               y.right = nd;
            } else y.left = nd;

        }
        size++;
        maxDepth = Math.max(maxDepth, nd.depth);
    }

    public T search(K key) {
        Node x = root;
        while(x != null) {
            int compResult = comp.compare(key, x.key);
            if (compResult == 0) return x.payload;
            else if (compResult < 0) x = x.left;
            else x = x.right;
        }
        return null;
    }


    public void printTree() {
       printNode(root);
    }

    public void printNode(Node nd) {
        if (nd == null) return;
        System.out.println("Node: " + nd + " depth = " + nd.depth + " p=" + nd.p );
        printNode(nd.left);
        printNode(nd.right);
    }

    public void rotate(Node node) {

        Node parent = node.p;

        if (node == parent.left) {
            parent.left = node.right;
            if (parent.left != null) parent.left.p = parent;
            node.right = parent;
            node.p = parent.p;
            parent.p = node;
            if (node.p != null) {
                if (node.p.left == parent) node.p.left = node;
                else node.p.right = node;
            }
        } else if (node == parent.right) {
            parent.right = node.left;
            if (parent.right != null) parent.right.p = parent;
            node.left = parent;
            node.p = parent.p;
            parent.p = node;
            if (node.p != null) {
                 if (node.p.left == parent) node.p.left = node;
                else node.p.right = node;
            }
        }  else {
            assert(false);
        }
        if (node.p == null) root = node;
    }

    public void splay(Node node) {
        while (node != root) {
            Node grandparent = node.p.p;
            System.out.println("Double-rotating: " + node + " parent=" + node.p + " grandp=" + grandparent);

            if (grandparent == null) {
                rotate(node);
            } else {
                if ((node == node.p.left && node.p == grandparent.left ) ||
                        (node == node.p.right && node.p == grandparent.right)) {
                    rotate(node.p);
                    rotate(node);
                } else {
                    rotate(node);
                    rotate(node);
                }
            }
        }
    }


    public class Node {
        Node p;

        public Node getLeft() {
            return left;
        }

        public Node getRight() {
            return right;
        }

        public K getKey() {
            return key;
        }

        Node left;
        Node right;

        K key;
        T payload;

        int depth;

        public T getValue() { return payload; }

        public String toString() {
            return key.toString();
        }

    }

}
